/* simpleldap.c
 *	SASL auxprop module for ldap authentification via cmusaslsecret* passwords
 * ORIGIN : modified giengerldap module of Pascal Gienger (http://southbrain.com/south)
 * AUTHOR : Samuel Behan <sam_(at)_dob_(dot)_sk> (renamed, fixed some bugs, code cleanup, added %u param, hex-to-raw user password conversion...)
 * LICENSE: Pascal Gienger (hopefully GNU GPLv3 or similar)
 *
 * SAMPLE SASL CONFIG:
 *	auxprop_plugin: eldap
 *	eldap_uri: ldap://localhost
 *	eldap_basedn: o=myorganization,c=DE
 *	#eldap_binddn: cn=smtpauth,o=myorganization,c=DE
 *	#eldap_bindpw: mysuperpassword
 *	eldap_search: (uid=%u)
 *	#eldap_bind_retries: 3
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <assert.h>

#define LDAP_DEPRECATED 1	/* don't understand why ldap_simple_bind_s is deprecated, but it is */
#include <ldap.h>

#include <sasl.h>
#include <saslutil.h>
#include <saslplug.h>
//#include "plugin_common.h"

#define MAXSEARCHLEN 1024
#define DEF_BIND_RETRIES 3

typedef struct ldapctx
{
	const char *uri;
	const char *basedn;
	const char *binddn;
	const char *bindpw;
	const char *attribute;
	const char *search;
	int bind_retries;
} ldapctx_t;

/* plugin name */
#ifndef _PLUGNAME
#define _PLUGNAME "eldap"
#else
#warning Compiling to be compatible with GIENGER ldap module.
#endif
static const char const *_plugname = _PLUGNAME;
static const int ldap_version = 3;

/* function to replace % params */
static char *str_replace(char *buf, const char const *str2, const char const *str1,
	    const char const *replacement, int buflen)
{
	unsigned int i, c, l, r;
	char *ptr = NULL, *ptra = NULL, *ptrc = NULL;
	const char emptystr[] = "";

	if(replacement == NULL)
		replacement = emptystr;

	c = buflen;
	l = strlen(str1);
	r = strlen(replacement);
	ptrc = (char *) str2;
	ptra = (char *) str2;
	memset(buf, 0, buflen);

	while((ptr = strstr(ptrc, str1)) != NULL)
	{
		i = (unsigned int) (ptr - ptrc);
		if (c > i)
		{
			strncat(buf, ptrc, i);
			c -= i;
		}
		if (c > r)
		{
			strncat(buf, replacement, c);
			c -= r;
		}
		ptrc = ptr + l;
	}
	strncat(buf, ptrc, c - 1);
	buf[buflen - 1] = '\0';
	return buf;
}

static void eldap_auxprop_lookup(void *glob_context,
			   sasl_server_params_t * sparams,
			   unsigned flags, const char *user, unsigned ulen)
{
	ldapctx_t *ldapctx = glob_context;
	char *username = NULL;
	char searchstring[MAXSEARCHLEN];
	char buffer[MAXSEARCHLEN];
	char **attrlist = NULL;
	int erg, i, retries = 0, seconds = 1;
	char *str = NULL;
	const struct propval *properties = NULL;
	LDAPMessage *ldresult = NULL, *ldresults = NULL;
	LDAP *ld = NULL;

	/* init */
	erg = ldap_initialize(&ld, ldapctx->uri);
	if (erg != LDAP_SUCCESS)
	{
		if(ld)
			ldap_unbind_s(ld);
		syslog(LOG_ERR | LOG_AUTH, "%s - ldap_initialize() failed with uri %s\n", _plugname, ldapctx->uri);
		return;
	}

	/* force version 3 */
	erg = ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &ldap_version);

	/* bind to ldap */
	while(retries++ < ldapctx->bind_retries)
	{
		erg = ldap_simple_bind_s(ld, (const char *) ldapctx->binddn,
				 (const char *) ldapctx->bindpw);
		if (erg != LDAP_SUCCESS && retries < ldapctx->bind_retries)
		{
			syslog(LOG_WARNING | LOG_AUTH, "%s - ldap_simple_bind_s() failed with '%s' (%d), sleeping %d secs (retry %d of %d)\n",
					_plugname, ldap_err2string(erg), erg, seconds, retries, ldapctx->bind_retries);
			sleep(seconds);
			seconds = (retries <= DEF_BIND_RETRIES ? retries * 2 : retries);
		}
	}

	/* test if connect successfull */
	if(erg != LDAP_SUCCESS)
	{
		ldap_unbind_s(ld);
		syslog(LOG_ERR | LOG_AUTH, "%s - ldap_simple_bind_s() failed with '%s' (%d), giving up\n",
					_plugname, ldap_err2string(erg), erg);
		return;
	}

#ifdef DEBUG
	syslog(LOG_DEBUG | LOG_AUTH, "%s connected to LDAP\n", _plugname);
#endif

	memset(searchstring, 0, MAXSEARCHLEN);
	username = sparams->utils->malloc(ulen + 1);
	assert(username != NULL);
	memset(username, 0, ulen + 1);
	memcpy(username, user, ulen);

	/* replace supplied username: %s */
	strncpy(searchstring, ldapctx->search, MAXSEARCHLEN - 1);
	str_replace(buffer, searchstring, "%s", username, MAXSEARCHLEN);
	strncpy(searchstring, buffer, MAXSEARCHLEN - 1);

	/* remove realm from username if present and replace %u */
	if((str = strstr(username, "@")) != NULL)
		str[0]	= '\0';
	str_replace(buffer, searchstring, "%u", username, MAXSEARCHLEN);
	strncpy(searchstring, buffer, MAXSEARCHLEN - 1);

	/* replace realm (if present) */
	if(sparams->user_realm != NULL)
	{
		str_replace(buffer, searchstring, "%r", sparams->user_realm, MAXSEARCHLEN);
		strncpy(searchstring, buffer, MAXSEARCHLEN - 1);
	}

	/* replace servern name (if present) */
	if(sparams->serverFQDN != NULL)
	{
		str_replace(buffer, searchstring, "%h", sparams->serverFQDN, MAXSEARCHLEN);
		strncpy(searchstring, buffer, MAXSEARCHLEN - 1);
	}

	/* propreties for cmusaslsecret-TYPE */
	properties = sparams->utils->prop_get(sparams->propctx);
	if(!properties)
	{
		ldap_unbind_s(ld);
		syslog(LOG_ERR | LOG_AUTH, "%s - prop_get() returned no properties, can not determine auth algorithm\n", _plugname);
		return;
	}
	for(i = 0; properties[i].name != NULL; i++);
	attrlist = sparams->utils->malloc((i + 1) * sizeof(char *));
	assert(attrlist != NULL);

	for(i = 0; properties[i].name; i++)
	{
		attrlist[i] = (char *) (properties[i].name + 1);
#ifdef DEBUG
		syslog(LOG_DEBUG | LOG_AUTH, "%s - adding attribute request %i: %s\n", _plugname, i, (char *) (properties[i].name + 1));
#endif
	}
	attrlist[i] = NULL;

#ifdef DEBUG
	syslog(LOG_DEBUG | LOG_AUTH, "%s - ldap search '%s'\n", _plugname, searchstring);
#endif
	ldresults = NULL;
	erg = ldap_search_ext_s(ld, ldapctx->basedn, LDAP_SCOPE_SUBTREE, searchstring, attrlist, 0, NULL, NULL, NULL, 1, &ldresults);
	free(attrlist);

	if(erg != LDAP_SUCCESS)
	{
		ldap_unbind_s(ld);
		if(!ldresults)
			ldap_msgfree(ldresults);
		syslog(LOG_ERR | LOG_AUTH, "%s - ldap_search_ext_s() failed with '%s' (%d)\n", _plugname, ldap_err2string(erg), erg);
		return;
	}

#ifdef DEBUG
	syslog(LOG_DEBUG | LOG_AUTH, "%s - ldap search returned %d entries\n", _plugname, ldap_count_entries(ld, ldresults));
#endif

	/* go trough results*/
	ldresult = ldap_first_entry(ld, ldresults);
	if(!ldresult)
	{
		ldap_unbind_s(ld);
		syslog(LOG_ERR | LOG_AUTH, "%s - ldap_first_entry() failed with '%s' (%d)\n", _plugname, ldap_err2string(erg), erg);
		return;
	}

	for(i = 0; properties[i].name; i++)
	{
		struct berval **attrlenarray = NULL;
		struct berval *attrlen = NULL;

//		syslog(LOG_DEBUG | LOG_AUTH, "Testing: %s", (char *) (properties[i].name + 1));
		attrlenarray = ldap_get_values_len(ld, ldresult, (char *) (properties [i].name + 1));
		if(attrlenarray)
		{
			unsigned int x = 0;
			unsigned int g = 0;
			char password[256];

			attrlen = attrlenarray[0];
			g = attrlen->bv_len;
			if(g > sizeof(password) - 1)
				g = sizeof(password) - 1;
			memset(password, 0, sizeof(password));
			memcpy(password, attrlen->bv_val, g);

			/* recode password from hex->raw */
			if(attrlen->bv_len == 32)
			{
				for(x = 0; x < attrlen->bv_len / 2; x++)
				{
					int c;

					sscanf(attrlen->bv_val + (x * 2), "%02x", &c);
					password[x] = c;
				}
				password[x] = 0;
			}

			sparams->utils->prop_erase(sparams->propctx, properties[i].name);
			sparams->utils->prop_set(sparams->propctx, properties[i].name, password, g);
			memset(password, 0, 255);
#ifdef DEBUG
			syslog(LOG_DEBUG | LOG_AUTH, "%s result: %s (len = %d)\n", _plugname, (char *)(properties[i].name +  1), g);
#endif
			ldap_value_free_len(attrlenarray);
		}
	}

	/* free */
	if (ldresults)
		ldap_msgfree(ldresults);

	/* final unbind */
	ldap_unbind_s(ld);
}

static void eldap_auxprop_free(void *glob_ctx, const sasl_utils_t *utils)
{
	if (glob_ctx != NULL)
		utils->free(glob_ctx);
}

static sasl_auxprop_plug_t eldap_auxprop_plugin = {
	0,				/* Features */
	0,				/* spare */
	NULL,				/* glob_context */
	eldap_auxprop_free,	/* auxprop_free */
	eldap_auxprop_lookup,	/* auxprop_lookup */
	"eldap",			 /* name */
	NULL				/* spare */
};

int eldap_auxprop_plug_init(const sasl_utils_t *utils,
			      int max_version,
			      int *out_version,
			      sasl_auxprop_plug_t **plug,
			      const char *_plugname)
{
	ldapctx_t p1, *p2;
	const char *s = NULL;

	memset(&p1, 0, sizeof(p1));

	/* required params */
	utils->getopt(utils->getopt_context, _plugname, _PLUGNAME "_uri", &p1.uri, NULL);
	utils->getopt(utils->getopt_context, _plugname, _PLUGNAME "_basedn", &p1.basedn, NULL);
	utils->getopt(utils->getopt_context, _plugname, _PLUGNAME "_search", &p1.search, NULL);
	utils->getopt(utils->getopt_context, _plugname, _PLUGNAME "_bind_retries", &s, NULL);
	if(!p1.uri || !p1.basedn || !p1.search)
	{
		syslog(LOG_ERR | LOG_AUTH, "%s - params required: " _PLUGNAME "_uri, "_PLUGNAME "_basedn, "_PLUGNAME "_search\n", _plugname);
		return SASL_BADPARAM;
	}

	/* convert & sanitize retries */
	if(s != NULL)
	{
		p1.bind_retries = atoi(s);
		if(p1.bind_retries <= 0 || p1.bind_retries > 100)
		{
			syslog(LOG_WARNING | LOG_AUTH, "%s - param " _PLUGNAME "_retries - invalid value '%s', using 3 as default\n", _plugname, s);
			p1.bind_retries	= DEF_BIND_RETRIES;
		}
	}
	else	p1.bind_retries = 1;

	/* optional */
	utils->getopt(utils->getopt_context, _plugname, _PLUGNAME "_binddn", &p1.binddn, NULL);
	utils->getopt(utils->getopt_context, _plugname, _PLUGNAME "_bindpw", &p1.bindpw, NULL);

	/* alloc mem for context & copy data */
	p2 = utils->malloc(sizeof(ldapctx_t));
	assert(p2 != NULL);
	if (!p2)
		return SASL_NOMEM;
	memcpy(p2, &p1, sizeof(ldapctx_t));

	/* assign our context */
	eldap_auxprop_plugin.glob_context = p2;
	*out_version = SASL_AUXPROP_PLUG_VERSION;
	*plug = &eldap_auxprop_plugin;

	/* ready to go */
	syslog(LOG_INFO | LOG_AUTH, "%s init ok (uri = %s, basedn = %s)\n", _plugname, p1.uri, p1.basedn);
	return SASL_OK;
}
